package core.protocols.p2p.chord;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import core.protocols.p2p.Node;
import core.tools.Range;
import java.util.NavigableMap;
import java.util.concurrent.ConcurrentSkipListMap;
/**
 * This abstract class represent a node in the jChord protocol
 * 
 * @author laurent.vanni@sophia.inria.fr - logNet team 2010 - INRIA
 *         Sophia-Antipolis - France
 *         
 * Modified by Giang Ngo, giang.ngo_hoang@inria.fr 
 * - logNet team 2011 - INRIA Sophia-Antipolis - France
 * 
 */
public abstract class AbstractChord implements IChord {

	/** the degree of the address space of the chord */
	public static int SPACESIZE = 25;

	/** represent this node */
	private Node thisNode;

	/** Node predecessor */
	private Node predecessor;

	/** Node successor */
	//private Node successor;
	
	private NavigableMap<Integer, SuccessorInfo> successorList;
	
	private int successorListSize = 8;
	
	/** Fingers table */
	private Node[] fingersTable = new Node[SPACESIZE];

	/** hash Table containing the keys i am responsible for */
	protected Map<Integer, String> table;

	/** index of the next finger */
	private int next = SPACESIZE;

	/** Time between each stabilization in SECONDS */

	/** Time between fix fingerstable in SECONDS */
	
	private int stabilizeInterval = 30000;
	
	/** The status of the node */
	private boolean alive = true;
	
	/**
	 * Initialize a chord node
	 * 
	 * @param ip
	 *            the address ip of the node
	 * @param id
	 *            the id of the node
	 * @param port
	 *            the port number of the node to listen request
	 */
	public void initialize(String ip, int id, int port) {
		thisNode = new Node(ip, id, port);
		
		
		
		//moi sua
		predecessor = null;
		
		successorList = new ConcurrentSkipListMap<Integer, SuccessorInfo>();
		
		addSuccessor(thisNode);
		//System.out.println("init, successor 0"+ getSuccessor(0).toString());
				
		table = new HashMap<Integer, String>();
		for (int i = 1; i < SPACESIZE; i++)
			fingersTable[i] = thisNode;
		checkStable();

	}

	// /////////////////////////////////////////// //
	// CHORD ALGORITHM
	// /////////////////////////////////////////// //
	/**
	 * @see core.protocols.p2p.chord.IChord#findSuccessor(int)
	 */
	public Node findSuccessor(int id) {
		//System.out.println("find succ, thisNode: "+getSuccessor(0));
		if (Range.inside(id, thisNode.getId() + 1, getSuccessor(0).getId())) {
			return getSuccessor(0);
		} else {
			Node pred = closestPrecedingNode(id);
			String succ = sendRequest(FINDSUCC + "," + id, pred);
		//	System.out.println("chord succ:"+succ);
			Node result = succ.equals("unreachable") ? thisNode : new Node(succ);
			//System.out.println("find succ, succ: "+result.getId());	
			if (succ.equals("unreachable"))
				handleFailedNode(pred);
			return result;
		}
	}

	/**
	 * @see core.protocols.p2p.chord.IChord#closestPrecedingNode(int)
	 */
	public Node closestPrecedingNode(int id) {

		fingersTable[1] = getSuccessor(0);
		
		for (int i = SPACESIZE - 1; i > 0; i--) {
			if ((fingersTable[i] != null) && (Range.inside(fingersTable[i].getId(), 
						thisNode.getId() + 1,id - 1)))
				return fingersTable[i];
		}
		return thisNode;
	}

	
	
	
	/**
	 * @see core.protocols.p2p.chord.IChord#join(Node)
	 */
	public void join(Node chord) {
		String succ = sendRequest(FINDSUCC + "," + thisNode.getId(), chord);
		
		if(!succ.equals("unreachable")){
			
			Node dest = new Node(succ);
			String res = sendRequest(JOIN + "," + thisNode, dest);
			if (!res.equals("unreachable"))
			{
				//System.out.println("Result from JOIN request: "+res.toString());
				//System.out.println("join addsuccessor");
				addSuccessor(dest);
				ArrayList<Node> successors = new ArrayList<Node>();
				String[] nodes = res.split(";");
				predecessor = new Node(nodes[0]);
				for (int i=0; i < nodes.length; i++)
					successors.add(new Node(nodes[i]));
		//		System.out.println("join ok");
				updateSuccessorList(successors);
			}
		} else {
			System.out.println("Fail to join: to" + chord.getIp() + ":" + chord.getPort()
					+" this Node:"+thisNode.toString());
		}
	}

	
	/**
	 * @see core.protocols.p2p.chord.IChord#stabilize()
	 */
	public void stabilize() {
		
		//System.out.println("run stabilize");
		if ((successorList.size() != 0) && !getSuccessor(0).equals(thisNode))
		{
			
			String pred = sendRequest(Integer.toString(GETPRED), getSuccessor(0));
	
			if (!pred.equals("unreachable"))
			{	
				if (!pred.equals("")) { // PRED != NULL
						Node x = new Node(pred);
						if (Range.inside(x.getId(), thisNode.getId() + 1,
								getSuccessor(0).getId() - 1)) {
							//System.out.println("stabilize1, addsuccessor");
							addSuccessor(x);
						}
				}
			
					Node dest = getSuccessor(0);	
					String res = sendRequest(NOTIF + "," + thisNode, dest);
					if (!res.equals("unreachable"))
					{
						//System.out.println("stabilize2, addsuccessor");
						addSuccessor(dest);
						ArrayList<Node> successors = new ArrayList<Node>();
						String[] nodes = res.split(";");
						for (int i=0; i < nodes.length; i++)
						{
							//System.out.println("add:"+new Node(nodes[i]).toString());
							successors.add(new Node(nodes[i]));
						}	
						//System.out.println("Update in Stabilize"+successors.toString());
						
						updateSuccessorList(successors);
					}
					else 
					{
						//System.out.println("notify call handle failed Node:"+getSuccessor(0).toString());
						handleFailedNode(getSuccessor(0));
					}	
			}
			else
				handleFailedNode(getSuccessor(0));

			}
			
		}



	/**
	 * @see core.protocols.p2p.chord.IChord#notify()
	 */
	public String notify(Node node) {
		if ((predecessor == null)
				|| (Range.inside(node.getId(), predecessor.getId() + 1,
						thisNode.getId() - 1)))
		{
			Node oldPredecessor = predecessor;
			predecessor = node;
			//sent to old predecessor, message: code,thisNode,newSucc
			if (oldPredecessor != null)
					sendRequest(NEWSUCC + "," + thisNode +","+ node, oldPredecessor);

		}			
		return getAllSuccessors();
	}
	
	
	
	/**
	 * Fix the fingers table
	 
	private void fixFingersTable() {
		//System.out.println("run stabilize");
		next--;
		if (next <1 )
			next = SPACESIZE - 1;
		fingersTable[next] = findSuccessor((thisNode.getId() + (int) Math.pow(
				2, next - 1))
				% (int) Math.pow(2, SPACESIZE - 1));
	}
*/
	
	private void fixFingersTable() {
		next--;
		if (next < 1)
			next = SPACESIZE -1;
		fingersTable[next] = findSuccessor((thisNode.getId() + (int) Math.pow(
				2, next - 1))
				% (int) Math.pow(2, SPACESIZE - 1));
		}

	
	
	public void fixPredecessor(){
			if (predecessor != null)
					if (!pingNode(predecessor))
						handleFailedNode(predecessor);
	}

	
	// /////////////////////////////////////////// //
	// ALGORITHM +
	// /////////////////////////////////////////// //
	
	//giang
	
	protected String joinReceiver(Node node)
	{
	//	System.out.println("chord"+thisNode.toString()+"Receive join from "+node.toString());
		Node oldPredecessor = predecessor;
		//getObjectOnJoin(node);
	
		//this node is only node on the ring
		if (oldPredecessor == null)
				oldPredecessor = thisNode;
		//message return: oldPredecessor;thisNode;successors
		String result = oldPredecessor.toString()+";"+getAllSuccessors();  
		
		//sent to old predecessor, message: code,thisNode,newSucc
		if (predecessor != null)
				sendRequest(NEWSUCC +","+ node, predecessor);
			
		predecessor = node;	
		
		if (successorList.size() == 1 && getSuccessor(0) == thisNode)
		{
			addSuccessor(node);
			//System.out.println("joinreceiver addsuccessor");
		}
		return result;
		
	}
	
	protected void newSuccHintReceiver(Node newSucc)
	{
	//	System.out.println("Receive newsucc hint: "+newSucc.toString());
		if (Range.inside(newSucc.getId(), thisNode.getId() + 1, 
				getSuccessor(0).getId() - 1))
			{
			//System.out.println("newsucchintreceiver addsuccessor");
				addSuccessor(newSucc);
				//System.out.println("Newsucchint, set succ to "+ newSucc.toString());
			}
		if (getSuccessor(0).getId() == newSucc.getId())
				stabilize();
	}

	
	/**
	 * Get the objects to the node responsible for
	 * 
	 * @param node
	 */
	protected void getObjectOnJoin(Node node) {
		Object[] keys = table.keySet().toArray();
		for (Object key : keys) {
			if (!Range.inside((Integer) key, node.getId(), getThisNode()
					.getId())) {
				sendRequest(PUT + "," + (Integer) key + "," + table.get(key), node);
				table.remove(key);
			}
		}
	}

	/**
	 * @see core.protocols.p2p.chord.IChord#kill()
	 */
	public void kill() {
		
	}

	
	// /////////////////////////////////////////// //
	// ABSTRACT METHODS
	// /////////////////////////////////////////// //
	/**
	 * forward a message to an other node
	 * 
	 * @param message
	 * @param destination
	 * @return the response to this message
	 */
	public abstract String sendRequest(String message, Node destination);

	// /////////////////////////////////////////// //
	// DISPLAY
	// /////////////////////////////////////////// //
	public String toString() {
		String res = "<NODE: " + thisNode.getId() + ", PRED: "
				+ (predecessor == null ? predecessor : predecessor.getId())
				+ ", SUCC: "
				+ (getSuccessor(0) == null ? getSuccessor(0) : getSuccessor(0).getId()) + "> ";
		res += "\n\tFingers Table: ";
		if (fingersTable[1] != null) {
			res += "[";
			for (int i = 1; i < SPACESIZE - 1; i++) {
				res += fingersTable[i].getId() + ", ";
			}
			res += fingersTable[SPACESIZE - 1].getId() + "]";
		} else {
			res += "null";
		}
		if (!table.isEmpty()) {
			res += "\n\n\tData Content : ";
			for (Map.Entry<Integer, String> entry : table.entrySet()) {
				res += "\n\t  [" + entry.getKey() + "] - ";
				res += entry.getValue().toString();
			}
		}
		res += "\n\n";
		return res;
	}

	// /////////////////////////////////////////// //
	// GETTER AND SETTER //
	// /////////////////////////////////////////// //
	/**
	 * @see core.protocols.p2p.chord.IChord#getThisNode()
	 */
	public Node getThisNode() {
		return thisNode;
	}

	/**
	 * @see core.protocols.p2p.chord.IChord#getPredecessor()
	 */
	public Node getPredecessor() {
		return predecessor;
	}

	/**
	 * Set the predecessor of the node
	 * 
	 * @param n
	 */
	public void setPredecessor(Node n) {
		predecessor = n;
	}

	/**
	 * Set the successor of the node
	 * 
	 * @param n
	 */

	
/*
	public Node getSuccessor() {
		return successor;
	}
*/	
	
	
	public void checkStable() {
		new Thread(new Runnable() {
			public void run() {
				while (alive) {
					try {
						stabilize();
						fixFingersTable();
						fixPredecessor();
						Thread.sleep(stabilizeInterval);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		}).start();
	}
	

	public void handleFailedNode(Node failedNode)
	{
		//System.out.println("handle failed node "+failedNode);
	   if ((predecessor != null) && failedNode == predecessor)
	        predecessor = null;
	   Node oldSuccessor = getSuccessor(0);
	   successorListHandleFailedNode(failedNode);
	   fingerTableHandleFailedNode(failedNode);

	   if ((predecessor != null) &&
		        oldSuccessor == predecessor) {
		        predecessor = null;
		    }	
	}
	
	public void successorListHandleFailedNode(Node failedNode)
	{
		//System.out.println("successorlist failed Node:"+failedNode.toString());
		Iterator iter = successorList.keySet().iterator();
		int key = -1;
		boolean stop = false;
		while (iter.hasNext() && !stop)
		{
			//System.out.println("successorlist failed Node:"+key+":"+failedNode.toString());
			key = (Integer)iter.next();
			if (failedNode.equals(successorList.get(key).successor))
			 stop = true;
		}
		
		if (key != -1) 
			successorList.remove(key);
		if (successorList.size() == 0)
			addSuccessor(thisNode);
	}
	
	
	public void fingerTableHandleFailedNode(Node failedNode)
	{
	//	System.out.println("fingertable failed Node:"+failedNode.toString());
		int size = fingersTable.length;
		for (int i = size-1; i >=0; i-- )
		{
			if ((fingersTable[i] != null) && (fingersTable[i].equals(failedNode)))
				fingersTable[i] = null;
		}
	}
	
	
	public Boolean pingNode(Node dest)
	{
		String res = sendRequest(Integer.toString(IChord.PING), dest);
		if ((res != null) && (res.equals(Integer.toString(IChord.PONG))))
			return true;
		else 
			return false;
	}
	
	
	
	public Node getSuccessor(int pos) {
		if (pos > successorListSize)
				System.out.println("Out of successorListSize");
		if (successorList.size() == 0)
		{
			System.out.println("getSuccessor null");
				return null;
		}		
		Iterator iter = successorList.keySet().iterator();
		for (int i=0; i< pos; i++)
		{
			iter.next();	
		}
		int key = (Integer)iter.next();
		return successorList.get(key).successor;
	}
	
	public String getAllSuccessors()
	{
		String successors = "";
		Iterator iter = successorList.keySet().iterator();
		while (iter.hasNext())
			successors = successors + ";"+ successorList.get(iter.next()).successor.toString();
		if (!successors.equals(""))
			successors = successors.substring(1);
		
		//System.out.println("all successor: "+successors);
		return successors;
	}
	
	
	
	public void updateSuccessorList(ArrayList<Node> newSuccessorList)
	{
				
		for (int i =0; i< newSuccessorList.size(); i++)
		{
			if (!Range.inside(newSuccessorList.get(i).getId(), thisNode.getId(), 
					getSuccessor(0).getId()))
			{
				//System.out.println("update successor list addsuccessor");
				addSuccessor(newSuccessorList.get(i));
			}
		}
		//System.out.println("before update");
		//traceSuccessorList();
		removeOldSuccessors();
		//traceSuccessorList();
		if (successorList.size() == 0) 
			System.out.println("no successor");
		
		
		
		for (int i=0; i< successorList.size() - successorListSize; i++)
			successorList.pollLastEntry();
		
		//traceSuccessorList();	
	}
	
	
	
	public void removeOldSuccessors()
	{
		ArrayList<Integer> keyToRemove = new ArrayList<Integer>();
		Iterator iter;
		iter = successorList.keySet().iterator();
		while(iter.hasNext())
		{
			int key = (Integer)iter.next();
			if (successorList.get(key).newEntry == false)
				keyToRemove.add(key);
			else 
					successorList.get(key).newEntry = false;

		}
		for (int i =0; i< keyToRemove.size(); i++)
			successorList.remove(keyToRemove.get(i));
		
	
	}
	
	

	public void addSuccessor(Node successor)
	{
		int distance = Range.distance(thisNode.getId()+1, successor.getId());
			
			if (successorList.containsKey(distance))
				successorList.remove(distance);
			
				SuccessorInfo successorInfo = new SuccessorInfo(successor, true);
				successorList.put(distance, successorInfo);
	
		
//		/traceSuccessorList();
	}

public class SuccessorInfo{
	public Node successor;
	public Boolean newEntry;
	public SuccessorInfo(Node successor, Boolean newEntry)
	{
		this.successor = successor;
		this.newEntry = newEntry;
	}
}

public void traceSuccessorList()
{
	if (successorList.size() > 0)
	{
		//System.out.println("successorList:");
		Iterator iter = successorList.keySet().iterator();
		while (iter.hasNext())
		{	
			int key = (Integer)iter.next();
			System.out.println("succ:"+key+":"+successorList.get(key).successor
					+":"+successorList.get(key).newEntry);
		}	
	}
	else 
		System.out.println("successorList size: " + 0);
	
}

}
