package experiments.current.node.chord;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import static java.util.concurrent.TimeUnit.SECONDS;
import core.protocols.p2p.Node;
import core.protocols.p2p.chord.AbstractChord;
import core.protocols.p2p.chord.IChord;
import core.protocols.transport.ITransport;
import core.protocols.transport.socket.request.RequestHandler;
import core.protocols.transport.socket.server.SocketImpl;
import core.tools.HashFunction;
import core.tools.Range;
import experiments.current.MessageCodes;
import experiments.current.Oracle;

/**
 * This is an implementation of a chord node
 * 
 * @author laurent.vanni@sophia.inria.fr - logNet team 2010 - INRIA
 *         Sophia-Antipolis - France
 * 
 * Modified by giang.ngo_hoang@inria.fr 
 * - logNet team 2011 - INRIA Sophia-Antipolis - France
 */
public class ChordNode extends AbstractChord {
	/** name of the service */
	private static SimpleDateFormat formater = new SimpleDateFormat(
	"dd/MM/yy_H:mm:ss");
	protected static String time = formater.format(new Date());
	public String overlayIntifier = "chordNetwork"; // use an unique ID is possible
	/** Transport protocol */
	protected ITransport transport;
	/** Hash function */
	private HashFunction h;
	/* just to debug */
	public static boolean debugMode = false;
	protected Node manager;
	//protected int delay = 60;
	protected int messageNumber = 0;
	protected long startTime;
	protected long endTime;
	ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(1);

	/**
	 * Default constructor
	 * 
	 * @param ip
	 * @param port
	 * @param overlayIntifier
	 *            the identifier of the chord network
	 */
	public ChordNode(String ip, int port, String overlayIntifier) {
		
		this.overlayIntifier = overlayIntifier;
		this.h = new HashFunction(overlayIntifier);
		// TRANSPORT LAYER BASED ON THE SOCKET IMPLEMENTATION
		transport = new SocketImpl(port, 10, RequestHandler.class.getName(),
				10, 1, 100, this);
		((SocketImpl) transport).launchServer();
		int id = h.SHA1ToInt(ip +  transport.getPort() + time);
		startTime = System.currentTimeMillis();
		//System.out.println("Create Chord"+ip+","+id+","+port+","+overlayIntifier);
		
		initialize(ip, id, transport.getPort());
	
		//	String key = String.valueOf(id);
		//String value = key;
	//	scheduler.schedule(new InitializeData(key, value), delay, SECONDS);
	}

	/**
	 * Constructor without overlayIntifier (it is auto-generate)
	 * 
	 * @param ip
	 * @param port
	 */
	public ChordNode(String ip, int port) {
		this(ip, port, "<" + ip + port + ">");
	}

	public ChordNode(String ip, int port, String overlayIntifier, String manager, int putDelay)
	{
		
		this(ip, port, overlayIntifier);
		this.manager = new Node(manager);
		String nId = String.valueOf(getThisNode().getId());	
		scheduler.schedule(new InitializeData(nId, nId), (long)putDelay, SECONDS);
	}
	

	/**
	 * @see core.protocols.p2p.chord.AbstractChord#sendRequest(String, Node)
	 */
	public String sendRequest(String message, Node destination) {
		messageNumber++;
		return transport.sendRequest(getIdentifier() + "," + message,
				destination);
	}

	/**
	 * @see core.protocols.p2p.chord.AbstractChord#join(String, int)
	 */
	public void join(String host, int port) {
		//System.out.println("chord bootstrap: "+host+","+port);
		Node chord = new Node(host, keyToH(host + port), port);
		join(chord);
	}

	/**
	 * @see core.protocols.p2p.IDHT#put(String, String)
	 */
	public void put(String key, String value) {
	//	System.out.println("Chord put: "+key+","+value+" at "+System.currentTimeMillis());
		int hKey;
		hKey = keyToH(key);
		put(hKey, value);
	}

	
	
	/**
	 * Recursive put
	 * 
	 * @param hKey
	 * @param value
	 */
	

	private void put(int hKey, String value) {
		if (Range.inside(hKey, getPredecessor().getId() + 1, getThisNode()
				.getId())) {
				table.put(hKey, value);
		} else {
			Node dest = null;
			if (getThisNode().equals(closestPrecedingNode(hKey)))
			{
				if (!getThisNode().equals(getSuccessor(0)))
					dest = getSuccessor(0);
				else 
					{
						//System.out.println("Can not put"+hKey);
						return;
					}
			}	
			else 
				dest =closestPrecedingNode(hKey);
		//	System.out.println("Send PUT"+hKey+" to:"+dest.toString());
			sendRequest(IChord.PUT + "," + hKey + "," + value, dest);
		}
	}


	/**
	 * iterative get
	 * 
	 * @param hKey
	 * @return the value associate to the key
	 */
/*	
	public String get(int hKey) {
		if (Range.inside(hKey, getPredecessor().getId() + 1, getThisNode()
				.getId())) {
			return table.get(hKey);
		} else {
			return sendRequest(IChord.GET + "," + hKey,
					closestPrecedingNode(hKey));
		}
	}
*/	
	//-------Iterative get---------------	

	/**
	 * iterative get
	 * 
	 * @param hKey
	 * @return the value associate to the key
	 */
	
	public String get(String key) {
		
		//System.out.println("Chord, getString key "+key);
		int count = 0;
		int hKey;
		String message = "";
		hKey = keyToH(key);	
		String shKey = String.valueOf(hKey);
		message = get(shKey, count);
		//System.out.println("Chord, get key:"+key+"Hkey:"+hKey+" result "+message);
		return message;
	}


	public String get(String shKey, int count) {
//		System.out.println("Chord, getH key "+hKey);
		int hKey = Integer.parseInt(shKey);
		if ((getPredecessor() != null) && Range.inside(hKey, getPredecessor().getId() + 1, getThisNode()
				.getId())) {
			return table.get(hKey)+","+count;
		}			
		
		Boolean cont = true;
		Node lastHop = null;		
		String result = "null";		
		Node nextHop = closestPrecedingNode(hKey);
		
		//System.out.println("closestPrecedingNode: "+nextHop.toString());
		
		if (getThisNode().equals(nextHop)) 
			lastHop = getSuccessor(0);
			
		
		if (lastHop == null)
			while (cont)
			{					
				//request form: 0,netId, 1GETcode,2key,3count,4srcNode
				count+=1;
				String getRequest = IChord.GETCLOSESTPRED + "," + hKey 
					+ ","+count +"," +getThisNode().toString();
				
				//String getRequest = IChord.GET + "," + hKey + ","+count +"," +getThisNode().toString();
				//System.out.println("nextHop:"+nextHop);
				
				String res = sendRequest(getRequest, nextHop);	
				//System.out.println("res:"+res);
				
				if (!res.equals("unreachable"))
				{	
					count +=1;
					if (res.length()>= 7){
						String code = res.substring(0,7);
						if (code.equals("LASTHOP"))
						{
							lastHop =  new Node(res.substring(7));
							cont =false;
						}
						else if (code.equals("NEXTHOP"))
						{
							nextHop = new Node(res.substring(7));		
						}
					}	
				}
				else 
					cont = false;
			}		
		
		if (lastHop == null)
			return result+","+count;
		else
		{	
			count+=1;
			String getRequest = IChord.GET + "," + hKey 
				+ ","+count +"," +getThisNode().toString();
			result = sendRequest(getRequest, lastHop);
			count+=1;
			return result+","+count;
		}	
	}

	public String getClosestPrecedingNode(int hKey)
	{
		Node nextHop = closestPrecedingNode(hKey);
		
		if (!getThisNode().equals(nextHop))
			return ("NEXTHOP"+nextHop.toString());
		else 
			return ("LASTHOP"+getSuccessor(0).toString());
	}
	
public String getReceiver(int hKey)
{
		return (table.get(hKey)); //DaTa 	
}
	
//-------end of iter get

public String pingReceiver()
{
	return Integer.toString(IChord.PONG);
}
/**
	 * @see core.protocols.p2p.IDHT#get(String)
	 */

	/**
	 * @see core.protocols.transport.IRequestHandler#handleRequest(String)
	 */
	public String handleRequest(String code) {
		if (debugMode) {
			System.out.println("\n** DEBUG: doStuff\n*\tcode: " + code);
		}
		//System.out.println("Chord receive message:" +code);
		String[] args = code.split(",");
		String result = "";
		if (args[0].equals(getIdentifier())) {
			int f = Integer.parseInt(args[1]);
			switch (f) {
			case IChord.GETPRED:
				if (getPredecessor() != null)
					result = getPredecessor().toString();
				break;
			case IChord.FINDSUCC:
				result = findSuccessor(Integer.parseInt(args[2])).toString();
				break;
			case IChord.GETCLOSESTPRED:
				result = getClosestPrecedingNode(Integer.parseInt(args[2])).toString();
				break;				
			case IChord.NOTIF:
				result = notify(new Node(args[2], Integer.parseInt(args[3]), Integer
						.parseInt(args[4])));
				break;			
			case IChord.JOIN:
				result = joinReceiver(new Node(args[2], Integer.parseInt(args[3]),
						Integer.parseInt(args[4])));
				break;
			case IChord.PUT:
				put(Integer.parseInt(args[2]), args[3]);
				break;
			case IChord.GET:
				result = getReceiver(Integer.parseInt(args[2]));
				break;
			case IChord.PING:
				result = pingReceiver();
				break;
			case IChord.SETPRED:
				setPredecessor(new Node(args[2], Integer.parseInt(args[3]),
						Integer.parseInt(args[4])));
				break;
			case IChord.NEWSUCC:
				newSuccHintReceiver(new Node(args[2], Integer.parseInt(args[3]),
						Integer.parseInt(args[4])));
				break;	
			// ORACLE TESTS
			case Oracle.PUT:
				put(args[2], args[3]);
				break;
			case Oracle.GET:
				result = get(args[2]);
				break;
			default:
				break;
			}
		} else if (args[0].equals("getIdentifier")) {
			return getIdentifier();
		}

		else if (args[0].equals(String.valueOf(MessageCodes.DELETE))) {
			result = deleteReceiver();
		}
		else if (args[0].equals(String.valueOf(MessageCodes.MANAGERCOL))) {
			result = deleteReceiver();
		}
		else if (args[0].equals(String.valueOf(MessageCodes.MANAGERGET))) {
			
			//System.out.println("Chord receive ManagerGET message");
			result = get(args[1]);
		}

				
		if (debugMode) {
			System.out.println("*\tresult: " + result
					+ "\n************************************");
		}
		return result;
	}
	
	@Override
	public String toString() {
		return getIdentifier() + " on " + getThisNode().getIp() + ":"
		+ getThisNode().getPort() + "\n" + super.toString();
	}

	/**
	 * @see core.protocols.p2p.IDHT#getTransport()
	 */
	public ITransport getTransport() {
		return transport;
	}

	/**
	 * @see core.protocols.p2p.IDHT#getIdentifier()
	 */
	public String getIdentifier() {
		return overlayIntifier;
	}

	/**
	 * @param key
	 * @return the hash value of the key
	 */
	public int keyToH(String key) { // A CHANGER!
		return h.SHA1ToInt(key);
	}
	
	public String deleteReceiver()
	{
		//System.out.println("Chord receive DELETE message");
		endTime = System.currentTimeMillis();
		int duration = Math.round((endTime - startTime)/1000);
		String message = "chord,"+messageNumber+","+String.valueOf(duration);
		//System.out.println("In Chord number of message:"+message);
		scheduler.schedule(new StopServer(), 10, SECONDS);
		return message;
		//return messageNumber,duration
	}
	
	
	
class StopServer implements Runnable
{
	public void run()
	{
		//System.out.println("ChordNode: "+ getThisNode().toString()+" killed");
		((SocketImpl)transport).stopServer();
		System.exit(0);
	}
}
class InitializeData implements Runnable
{
	String key;
	String value;

	public InitializeData(String key, String value)
	{
		this.key = key;
		this.value = value;
	}
	public void run()
	{
	//	System.out.println("Put"+key+":"+value+"to the chord net");
		put(key, value);
	}
}
	
}
