/**
 * 
 * @author giang.ngo_hoang@inria.fr - logNet team 2011 - INRIA
 *         Sophia-Antipolis - France
 */

package operations;

import static java.util.concurrent.TimeUnit.SECONDS;
import java.util.ArrayDeque;
import java.util.Iterator;
import java.util.Random;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import manager.Node;
import manager.AliveNodes;
import manager.Configuration;
import manager.Machines;
import manager.TimeCheck;
import tools.ParetoDistribution;
import transport.ITransport;

public class CreateOverlay implements Runnable
{
	
	protected String nodeType, netId;
	public static int nodeCounter;
	protected double initialDeviation;
	protected ITransport transport;
	protected Machines machines;
	protected AliveNodes aliveNodes;
	protected Configuration config;
	public Random randomGenerator;
	protected Node manager;
	protected ParetoDistribution pareto; 
	protected TimeCheck timeCheck;
	ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(1);
	
	// simulationTime =initTime + transition time+ put time + put/get time
	
	public CreateOverlay(String nodeType, String netId, Configuration config, 
			AliveNodes aliveNodes, ITransport transport, 
			Machines machines, Node manager, TimeCheck timeCheck)
	{
		this.nodeType = nodeType;
		this.netId = netId;
		this.config = config;
		this.aliveNodes = aliveNodes;
		this.transport = transport;
		this.machines = machines;
		this.manager = manager;		
		randomGenerator = new Random();
		pareto = new ParetoDistribution();
		this.timeCheck = timeCheck;
	}
	

	
	public void run()
	{
		if (config.churn == true)
			createChurnOverlay();
		else
			createStaticOverlay();
	}

	
	public void createChurnOverlay()
	{
		double initialMean = config.initPhaseCreationInterval;
		initialDeviation = initialMean / 3;
		
		int averageAliveNodes;
		if (nodeType.equals("SYNAPSE")) 
		{	
			averageAliveNodes = config.numberOfSynapsePerCreationThread;
		//	System.out.println("synapse:"+ config.numberOfSynapsePerCreationThread);
		}	
		else 
			averageAliveNodes = config.numberOfDataNodePerOverlay;
		int liveNodes = 0;
	    double sum_l_i = 0;  
	    ArrayDeque<node_stat> node_stats = new ArrayDeque<node_stat>();
	
	    for (int i = 0; liveNodes < averageAliveNodes; i++) {

	    	double nodeLifetimeMean = pareto.individualMeanTime(config.lifetimeMean); // lifemean of each node
	        double nodeDeadtimeMean = pareto.individualMeanTime(config.deadtimeMean); //deadmean of each node
	        
	        sum_l_i += 1.0/(nodeLifetimeMean + nodeDeadtimeMean);
	        
	        double nodeAvailability = nodeLifetimeMean/(nodeLifetimeMean + nodeDeadtimeMean); // ty le availability
	        node_stat nodeStat = new node_stat(nodeLifetimeMean, nodeDeadtimeMean, pareto.uniform(0, 1) < nodeAvailability);  //one specfic node
	 
	        if (nodeStat.alive) {
	            liveNodes++;  // live nodes are created in INIT phase
	        }
	        
	        node_stats.add(nodeStat);
	    }

	    // compute "stretch" factor to reach the configured average lifetime
	    // this is neccessary as "individual" lifetime mean has to be bigger than
	    // "global" lifetime mean, as short-lived nodes will factor in more often
	    double mean_life = 0;
	    int numNodes = node_stats.size(); 
	    Iterator<node_stat> iter = node_stats.iterator();
	    
	    while(iter.hasNext()){
	        node_stat stat = iter.next();  
	        mean_life += stat.l/( (stat.l + stat.d) * sum_l_i );
	    }
	    double stretch = config.lifetimeMean/mean_life;
	    liveNodes = 0; 

	    
	    
	    // schedule creation for all (alive or dead) nodes
		//System.out.println("numNodes: "+numNodes);
	    for( int i = 0; i < numNodes; ++i ){
	    
	        node_stat stat = node_stats.pollFirst(); //lay ra phan tu dau tien va xoa trong queue
	        stat.l *= stretch;
	        stat.d *= stretch;

	        if( stat.alive ){
			//initialMean*liveNodes: thoi gian create node hien tai
			//scheduleTime: time of creation, //create node at scheduleTime (time corresponding to each node)
			//initFinishedTime - scheduleTime + residualLifetime(stat.l) : life time cho node hien tai
	      
	        	
	        double scheduleTime = pareto.truncnormal(initialMean*liveNodes, initialDeviation, 0);
	        //debug
	        //System.out.println("schedule Alive Node: Creation time: "+scheduleTime);
	       
	        scheduler.schedule(new CreateNode(scheduler, config.initTime - scheduleTime + 
	        	pareto.residualLifetime(stat.l), stat.l, stat.d, machines, nodeType, 
	        	netId, manager, config, aliveNodes, transport, timeCheck), (long)scheduleTime, SECONDS);
	        // lifetime = initFinishedTime - scheduleTime + residualLifetime(stat.l) => time for InitPhase - time of creation + lifetime con lai

	            liveNodes++;
	        } else {
	           
	        	double scheduleTime = config.initTime + pareto.residualLifetime(stat.d);
	        	//debug
	        	//System.out.println("schedule, Dead Node: Creation time: "+scheduleTime);
	        	
	        	scheduler.schedule(new CreateNode(scheduler, pareto.individualLifetime(stat.l), stat.l, 
	        			stat.d, machines, nodeType, netId, manager, config, aliveNodes, transport, timeCheck), 
	        			(long)scheduleTime, SECONDS);
	        	
	        	
	        	
	        	
	        	
	       }
	    }    	
		
	}
	public void createStaticOverlay()
	{
		int averageAliveNodes;
		if (nodeType.equals("SYNAPSE")) 
			averageAliveNodes = config.numberOfSynapsePerCreationThread;
		else 
			averageAliveNodes = config.numberOfDataNodePerOverlay;
		 
		for (int i = 0; i < averageAliveNodes; i++) {
			long delay = (long)(config.initPhaseCreationInterval*(i+1));
			//System.out.println("Create static node at "+delay);
			
			scheduler.schedule(new CreateNode(scheduler, 0, 0, 0, machines, nodeType, 
		        	netId, manager, config, aliveNodes, transport, timeCheck), 
		        	delay, SECONDS);
		}
		
	}
	
	//---------------node_stat class ---------------
	class node_stat
	{
		double l;
		double d;
		boolean alive;
		public node_stat(double nodeLifetimeMean, double nodeDeadtimeMean, boolean status)
		{
			l = nodeLifetimeMean;
			d = nodeDeadtimeMean;
			alive = status;
		}
	}
	
	
	
}